package ppx_typed_fields

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module type As_applicative.S_for_other_map

This module is basically "Applicative", but with an additional type that can be used to translate to the applicative type.

Typically these types will be something like

  type 'a t = 'a Value.t
  type 'a s = 'a Form.t

  val translate: 'a Form.t Value.t -> 'a Form.t Value.t

translate probably doesn't need to do any work. It just exists to expose a type equality between 'a Data.t and 'a s t.

type 'a t
val map : 'a t -> f:('a -> 'b) -> 'b t
val all : 'a t list -> 'a list t
type 'a s
val translate : 'a Data.t -> 'a s t