package ppx_typed_fields

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Typed_fields_libSource

Sourcemodule type S = sig ... end
Sourcemodule type S1 = sig ... end
Sourcemodule type S2 = sig ... end
Sourcemodule type S3 = sig ... end
Sourcemodule type S4 = sig ... end
Sourcemodule type S5 = sig ... end
Sourcemodule S_of_S1 (M : S1) (T : Base.T) : S with type 'a t = (T.t, 'a) M.t and type derived_on = T.t M.derived_on
Sourcemodule S_of_S2 (M : S2) (T1 : Base.T) (T2 : Base.T) : S with type 'a t = (T1.t, T2.t, 'a) M.t and type derived_on = (T1.t, T2.t) M.derived_on
Sourcemodule S_of_S3 (M : S3) (T1 : Base.T) (T2 : Base.T) (T3 : Base.T) : S with type 'a t = (T1.t, T2.t, T3.t, 'a) M.t and type derived_on = (T1.t, T2.t, T3.t) M.derived_on
Sourcemodule S_of_S4 (M : S4) (T1 : Base.T) (T2 : Base.T) (T3 : Base.T) (T4 : Base.T) : S with type 'a t = (T1.t, T2.t, T3.t, T4.t, 'a) M.t and type derived_on = (T1.t, T2.t, T3.t, T4.t) M.derived_on
Sourcemodule S_of_S5 (M : S5) (T1 : Base.T) (T2 : Base.T) (T3 : Base.T) (T4 : Base.T) (T5 : Base.T) : S with type 'a t = (T1.t, T2.t, T3.t, T4.t, T5.t, 'a) M.t and type derived_on = (T1.t, T2.t, T3.t, T4.t, T5.t) M.derived_on
Sourcemodule Singleton (T : Base.T) : sig ... end
Sourcemodule Singleton1 (T1 : Base.T1) : sig ... end
Sourcemodule Singleton2 (T2 : Base.T2) : sig ... end
Sourcemodule Singleton3 (T3 : Base.T3) : sig ... end
Sourcemodule Singleton4 (T4 : sig ... end) : sig ... end
Sourcemodule Singleton5 (T5 : sig ... end) : sig ... end
Sourcemodule Unit : sig ... end

This is a convenient module for deriving typed_fields on unit, which you can conceptually think of as a record with no fields. OCaml does not support actual record types with no fields.

Sourcemodule Common : sig ... end
Sourcemodule Private : sig ... end